# -*- coding: utf-8 -*-
"""This module contains the controller classes of the application."""

# symbols which are imported by "from pagedemo2.controllers import *"
__all__ = ['Root']

# standard library imports
# import logging
import sys
from operator import itemgetter
from random import normalvariate, sample

# third-party imports
from cherrypy import request, response
from turbogears import (controllers, expose, flash, identity, paginate,
    redirect, url, validate, validators, visit)
from turbogears.widgets import PaginateDataGrid

# project specific imports
from pagedemo2.model import Group, User, session
from pagedemo2.lib.widgets import ImageLink
from pagedemo2.lib.names import generate_groups, generate_users

# log = logging.getLogger("pagedemo2.controllers")

# Grids for the User Management demo:

sortable = dict(sortable=True)

group_grid = PaginateDataGrid(name='Group',
    fields=[
        ('ID', 'group_id', sortable),
        ('Groupname', 'group_name', sortable),
        ('Full Name', 'display_name', sortable),
        ('Members', 'num_members', sortable),
        ('Eldest', 'eldest.display_name', sortable),
        ('', lambda group: ImageLink(link=url('group',
            group_id=group.group_id, query=request.query_string))())
    ])

user_grid = PaginateDataGrid(name='User',
    fields=[
        ('ID', 'user_id', sortable),
        ('Username', 'user_name', sortable),
        ('First Name', 'first_name', sortable),
        ('Last Name', 'last_name', sortable),
        ('Age', 'age',  sortable),
        ('Email Address', 'email_address', sortable),
        ('Password', 'password', sortable),
        ('Groups', 'num_groups', sortable),
        ('', lambda user: ImageLink(link=url('user',
            user_id=user.user_id, query=request.query_string))())
    ])


# Data and grid for the Python Modules demo:

modules = sys.modules
modules = [modules[name] for name in sorted(modules)
    if not name.startswith('_') and not '._' in name
        and modules[name]
        and (getattr(modules[name], '__doc__', '') or '').strip()]

def summary(s, max=76):
    s = s.lstrip().split('\n', 1)
    s, t = s[0].rstrip(), len(s) > 1 and s[1][:1] != '\n'
    if len(s) > max:
        s, t = s[:max-4].rsplit(' ', 1)[0], True
    if t:
        s += ' ...'
    return s

modules_grid = PaginateDataGrid(name='Python Modules',
    fields=[
        ('Name', '__name__'),
        ('Content', lambda module: summary(module.__doc__))
    ])


# Grid for the Fibonacci Number demo:

fibonacci_grid = PaginateDataGrid(name='Fibonacci Numbers',
    fields=[
        ('n', itemgetter(0)),
        ('F(n)', itemgetter(1))
    ])


# The TurboGears paginate demo 2 controller:

class Root(controllers.RootController):
    """The root controller of the application."""

    @expose(template="pagedemo2.templates.index")
    def index(self):
        return dict()

    @expose(template="pagedemo2.templates.grid")
    @paginate('data', limit=14, default_order='group_name')
    def groups(self):
        groups = Group.query
        count = groups.count()
        return dict(grid=group_grid, data=groups, count=count)

    @expose(template="pagedemo2.templates.view")
    @validate(validators=dict(group_id=validators.Int(not_empty=True),
        query = validators.UnicodeString()))
    def group(self, group_id, query=None):
        query = query and '?' + query or ''
        return dict(data=Group.query.get(group_id), backlink=url('groups' + query))

    @expose(template="pagedemo2.templates.grid")
    @paginate('data', default_order='user_name')
    def users(self):
        users = User.query
        count = users.count()
        return dict(grid=user_grid, data=users, count=count)

    @expose(template="pagedemo2.templates.users")
    @paginate('users', max_pages=None, default_order='user_name')
    def users2(self):
        users = User.query
        count = users.count()
        return dict(users=users, count=count)

    @expose(template="pagedemo2.templates.view")
    @validate(validators=dict(user_id=validators.Int(not_empty=True),
        query = validators.UnicodeString()))
    def user(self, user_id, query=None):
        query = query and '?' + query or ''
        return dict(data=User.query.get(user_id), backlink=url('users' + query))

    @expose()
    def populate(self):
        if User.query.first():
            flash("The tables are already populated.")
        else:
            users = []
            for user in generate_users():
                users.append(User(**user))
            groups = []
            for group in generate_groups():
                groups.append(Group(**group))
            for user in users:
                for group in sample(groups,
                        min(10, max(1, int(round(normalvariate(5, 3)))))):
                    user.groups.append(group)
            flash("The tables have been populated now.")
        redirect("/")

    @expose()
    def truncate(self):
        if User.query.first():
            for group in Group.query:
                group.users = []
                session.delete(group)
            for user in User.query:
                session.delete(user)
            flash("The tables have been truncated now.")
        else:
            flash("The tables are already truncated.")
        redirect("/")

    @expose(template="pagedemo2.templates.grid")
    @paginate('data', default_order='__name__')
    def modules(self):
        return dict(grid=modules_grid, data=modules, count=len(modules))

    @expose(template="pagedemo2.templates.modules")
    @paginate('modules', limit=1, max_pages=1)
    def modules2(self):
        return dict(modules=modules)

    @expose(template="pagedemo2.templates.grid")
    @paginate('data', max_pages=10)
    def fibonacci(self):
        F = [0, 1]
        while len(F) < 300:
            F.append(F[-1] + F[-2])
        return dict(grid=fibonacci_grid, data=enumerate(F), count=len(F))

    @expose(template="pagedemo2.templates.login")
    def login(self, forward_url=None, *args, **kw):
        """Show the login form or forward user to previously requested page."""

        if forward_url:
            if isinstance(forward_url, list):
                forward_url = forward_url.pop(0)
            else:
                del request.params['forward_url']

        new_visit = visit.current()
        if new_visit:
            new_visit = new_visit.is_new

        if (not new_visit and not identity.current.anonymous
                and identity.was_login_attempted()
                and not identity.get_identity_errors()):
            redirect(forward_url or '/', kw)

        if identity.was_login_attempted():
            if new_visit:
                msg = _(u"Cannot log in because your browser "
                         "does not support session cookies.")
            else:
                msg = _(u"The credentials you supplied were not correct or "
                         "did not grant access to this resource.")
        elif identity.get_identity_errors():
            msg = _(u"You must provide your credentials before accessing "
                     "this resource.")
        else:
            msg = _(u"Please log in.")
            if not forward_url:
                forward_url = request.headers.get("Referer", "/")

        # we do not set the response status here anymore since it
        # is now handled in the identity exception.
        return dict(logging_in=True, message=msg,
            forward_url=forward_url, previous_url=request.path_info,
            original_parameters=request.params)

    @expose()
    def logout(self):
        """Log out the current identity and redirect to start page."""
        identity.current.logout()
        redirect("/")
